# Discussion

## Task 1

The solution is incorrect because for any given (sub)tree it does not check **all** the values 
in the left/right subtree to be less/greater than value at root; it only checks the children. 

For example, the solution fails for the following case:

```text
      5
    /   \
   1     6
    \   / \
     7 3   8

Input: [5, 1, 6, null, 7, 3, 8]
Output: false
```

The test fails because the solution only checks `1 < 5` and fails to check for `7 < 5`.

## Task 2

If we always get head, we will do pre-order traversal.

If we always get tail when deciding to visit root, but get head when deciding to visit
left/right child, we will do post-order traversal.

It is possible to do in-order traversal if the in-order traversal is the same as either pre- or post-order.

Example where pre-order is the same as in-order

```text
     1
    / \
 null  2
      / \
   null  3
        / \
         ...
```

Example where in-order is the same as post-order

```text
     3
    / \
   2  null
  / \
 1  null
/ \
...
```


It is possible to get level-order traversal if the level-order is the same as pre-order or in-order 
(e.g. left/right-skewed binary tree where the pre- or post-order will be the same as in-order).

Example where level-order will give us the same as pre-order:

```text
      2
    /   \
  1       6
         / 
       4
      / \
     3   5      
```

## Task 3

**Please (very briefly) explain it to the student why this strategy will not work and, 
what is their misconception. Write your answer in the README.md file.**

In a heap, the highest (or lowest) priority element is always stored at the root (front of 
the rank array representation). However, a heap is not a sorted structure; so the last element
is not necessarily the "worst" element. More so in the rank array representation, the ordering
of the elements correspond to the structure of the tree and not their values.

**Describe the running time of your implementation for `best`, `worst`, `empty`, `insert`, 
`removeBest` and `removeWorst`.**

* `best`, `worst` and `empty` operations are $O(1)$ since they invoke constant-time operations 
  of `java.util.PriorityQueue`: `peek` and `isEmpty`
  
* `insert` is $O(\lg n)$ since it invokes logarithmic-time operation of `java.util.PriorityQueue`: 
  `add`
  
* `removeBest` and `removeWorst` operations are $O(n)$ since they invoke linear-time operation of
  `java.util.PriorityQueue`: `remove(Object)`

