package task3;

import exceptions.EmptyException;
import java.util.Comparator;

/**
 * An implementation of PriorityQueue with internally uses two java.util.PriorityQueue.
 *
 * @param <T> Element type.
 */
public class DoublePriorityQueueSln<T extends Comparable<T>> implements PriorityQueue<T> {

  private static final int INITIAL_CAPACITY = 11; // initial capacity of java.util.PriorityQueue
  private final java.util.PriorityQueue<T> bestHeap;
  private final java.util.PriorityQueue<T> worstHeap;

  /**
   * Construct a PriorityQueue.
   */
  public DoublePriorityQueueSln() {
    // The definition of "best" is in java.util.PriorityQueue is the opposite (reverse)
    // of how best is defined in task3.PriorityQueue:
    //  - best in java.util.PriorityQueue means the smallest
    //  - best in task3.PriorityQueue means the largest
    bestHeap = new java.util.PriorityQueue<>(INITIAL_CAPACITY, new ReverseNaturalOrdering());
    worstHeap = new java.util.PriorityQueue<>();
  }

  private class ReverseNaturalOrdering implements Comparator<T> {
    @Override
    public int compare(T s1, T s2) {
      return s2.compareTo(s1);
    }
  }

  /**
   * Construct a PriorityQueue with a given Comparator.
   *
   * @param comparator the comparator used to order the elements in this queue.
   */
  public DoublePriorityQueueSln(Comparator<T> comparator) {
    worstHeap = new java.util.PriorityQueue<>(INITIAL_CAPACITY, comparator);
    bestHeap = new java.util.PriorityQueue<>(INITIAL_CAPACITY, new ReverseComparator(comparator));
  }

  @Override
  public void insert(T t) {
    bestHeap.add(t);
    worstHeap.add(t);
    checkInvariant();
  }

  private class ReverseComparator implements Comparator<T> {
    private Comparator<T> comparator;

    ReverseComparator(Comparator<T> comparator) {
      this.comparator = comparator;
    }

    @Override
    public int compare(T s1, T s2) {
      return comparator.compare(s2, s1);
    }
  }

  @Override
  public void removeBest() throws EmptyException {
    checkPreCondition();
    T best = bestHeap.poll();
    worstHeap.remove(best);
    checkInvariant();
  }

  @Override
  public T best() throws EmptyException {
    checkPreCondition();
    return bestHeap.peek();
  }

  @Override
  public void removeWorst() throws EmptyException {
    checkPreCondition();
    T worst = worstHeap.poll();
    bestHeap.remove(worst);
    checkInvariant();
  }

  @Override
  public T worst() throws EmptyException {
    checkPreCondition();
    return worstHeap.peek();
  }

  @Override
  public boolean empty() {
    return bestHeap.isEmpty();
  }

  private void checkInvariant() {
    assert bestHeap.size() == worstHeap.size();
  }

  private void checkPreCondition() {
    if (empty()) {
      throw new EmptyException();
    }
  }
}
