package grade;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import task2.BinarySearchTree;
import task2.BinarySearchTreeSln;
import task2.Coin;
import org.junitpioneer.jupiter.StdIo;
import org.junitpioneer.jupiter.StdOut;
import java.util.Random;

import static org.junit.jupiter.api.Assertions.assertEquals;

public class BinarySearchTreeSlnTest {
  protected BinarySearchTree bst;

  protected BinarySearchTree createBst() {
    return new BinarySearchTreeSln();
  }

  private static class AlwaysTrueRandom extends Random {
    @Override
    public boolean nextBoolean() {
      return true;
    }
  }

  private static class FalseTrueRandom extends Random {
    private static boolean state = false;
    @Override
    public boolean nextBoolean() {
      state = !state;
      return !state;
    }
  }

  @Test
  @StdIo
  @DisplayName("[1] Generate pre-order traversal")
  void preOrder(StdOut out) {
    bst = createBst();
    /*
     *       5
     *     /   \
     *    1     7
     *     \   / \
     *      4 6   8
     */
    bst.insert(5);
    bst.insert(1);
    bst.insert(7);
    bst.insert(4);
    bst.insert(6);
    bst.insert(8);
    Coin.rand = new AlwaysTrueRandom();
    bst.traverse();
    String[] output = out.capturedLines();
    assertEquals("5 1 4 7 6 8 ", output[0]);
  }

  @Test
  @StdIo
  @DisplayName("[1] Generate post-order traversal")
  void postOrder(StdOut out) {
    bst = createBst();
    /*
     *       5
     *     /   \
     *    1     7
     *     \   / \
     *      4 6   8
     */
    bst.insert(5);
    bst.insert(1);
    bst.insert(7);
    bst.insert(4);
    bst.insert(6);
    bst.insert(8);
    Coin.rand = new FalseTrueRandom();
    bst.traverse();
    String[] output = out.capturedLines();
    assertEquals("4 1 6 8 7 5 ", output[0]);
  }

  @Test
  @StdIo
  @DisplayName("[1] Generate in-order traversal that is similar to pre-order")
  void preOrderSameAsInOrder(StdOut out) {
    bst = createBst();
    /*
     *      1
     *     / \
     *  null  2
     *       / \
     *    null  3
     *         / \
     *          ...
     */
    bst.insert(1);
    bst.insert(2);
    bst.insert(3);
    Coin.rand = new AlwaysTrueRandom();
    bst.traverse();
    String[] output = out.capturedLines();
    assertEquals("1 2 3 ", output[0]);
  }

  @Test
  @StdIo
  @DisplayName("[1] Generate in-order traversal that is similar to post-order")
  void postOrderSameAsInOrder(StdOut out) {
    bst = createBst();
    /*
     *      3
     *     / \
     *    2  null
     *   / \
     *  1  null
     * / \
     * ...
     */
    bst.insert(3);
    bst.insert(2);
    bst.insert(1);
    Coin.rand = new FalseTrueRandom();
    bst.traverse();
    String[] output = out.capturedLines();
    assertEquals("1 2 3 ", output[0]);
  }

  @Test
  @StdIo
  @DisplayName("[1] Generate level-order traversal that is similar to pre-order")
  void preOrderSameAsLevelOrder(StdOut out) {
    bst = createBst();
    /***
     *       2
     *     /   \
     *   1       6
     *          /
     *        4
     *       / \
     *      3   5
     */
    bst.insert(2);
    bst.insert(1);
    bst.insert(6);
    bst.insert(4);
    bst.insert(3);
    bst.insert(5);
    Coin.rand = new AlwaysTrueRandom();
    bst.traverse();
    String[] output = out.capturedLines();
    assertEquals("2 1 6 4 3 5 ", output[0]);
  }
}
