package grade;

import exceptions.EmptyException;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import task2.Coin;
import task3.PriorityQueue;

import java.util.Collections;
import java.util.Comparator;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

/**
 * Testing implementations of the PriorityQueue interface.
 */
public abstract class PriorityQueueTest {
  private PriorityQueue<Integer> unit;

  protected abstract PriorityQueue<Integer> createUnit();

  protected abstract PriorityQueue<Integer> createUnit(Comparator<Integer> comp);

  @BeforeEach
  public void setupTests() {
    unit = this.createUnit();
  }

  @Test
  @DisplayName("[0.5] PQ empty after construction")
  public void newQueueEmpty() {
    assertTrue(unit.empty());
  }

  @Test
  @DisplayName("[0.5] best() on empty PQ throws EmptyException")
  public void newQueueNoBest() {
    try {
      unit.best();
      fail("Exception was not thrown where it was expected");
    } catch (EmptyException ex) {
      return;
    }
  }

  @Test
  @DisplayName("[0.5] worst() on empty PQ throws EmptyException")
  public void newQueueNoWorst() {
    try {
      unit.worst();
      fail("Exception was not thrown where it was expected");
    } catch (EmptyException ex) {
      return;
    }
  }

  @Test
  @DisplayName("[0.5] removeBest() on empty PQ throws EmptyException")
  public void newQueueNoRemoveBest() {
    try {
      unit.removeBest();
      fail("Exception was not thrown where it was expected");
    } catch (EmptyException ex) {
      return;
    }
  }

  @Test
  @DisplayName("[0.5] removeWorst() on empty PQ throws EmptyException")
  public void newQueueNoRemoveWorst() {
    try {
      unit.removeWorst();
      fail("Exception was not thrown where it was expected");
    } catch (EmptyException ex) {
      return;
    }
  }

  @Test
  @DisplayName("[0.5] PQ not empty after insertion")
  public void insertNotEmpty() {
    unit.insert(1);
    assertFalse(unit.empty());
  }

  @Test
  @DisplayName("[0.5] PQ best() works: default Comparator")
  public void insertBestInOrder() {
    unit.insert(1);
    assertEquals(1, unit.best().intValue());

    unit.insert(3);
    assertEquals(3, unit.best().intValue());

    unit.insert(2);
    assertEquals(3, unit.best().intValue());
  }

  @Test
  @DisplayName("[0.5] PQ worst() works: default Comparator")
  public void insertWorstInOrder() {
    unit.insert(3);
    assertEquals(3, unit.worst().intValue());

    unit.insert(1);
    assertEquals(1, unit.worst().intValue());

    unit.insert(2);
    assertEquals(1, unit.worst().intValue());
  }

  @Test
  @DisplayName("[0.5] PQ best() works: non-default Comparator")
  public void insertBestReverseOrder() {
    unit = createUnit(Collections.reverseOrder());

    unit.insert(1);
    assertEquals(1, unit.best().intValue());

    unit.insert(3);
    assertEquals(1, unit.best().intValue());

    unit.insert(2);
    assertEquals(1, unit.best().intValue());
  }

  @Test
  @DisplayName("[0.5] PQ worst() works: non-default Comparator")
  public void insertWorstReverseOrder() {
    unit = createUnit(Collections.reverseOrder());

    unit.insert(3);
    assertEquals(3, unit.best().intValue());

    unit.insert(1);
    assertEquals(1, unit.best().intValue());

    unit.insert(2);
    assertEquals(1, unit.best().intValue());
  }

  @Test
  @DisplayName("[0.5] PQ removeBest() works: default Comparator")
  public void insertRemoveBestInOrder() {
    int[] nums = {5, 4, 3, 2, 1, 0};
    for (int i = nums.length - 1; i >= 0; i--) {
      unit.insert(nums[i]);
    }

    for (int n : nums) {
      assertEquals(n, unit.best().intValue());
      unit.removeBest();
    }
  }

  @Test
  @DisplayName("[0.5] PQ removeWorst() works: default Comparator")
  public void insertRemoveWorstInOrder() {
    int[] nums = {0, 1, 2, 3, 4, 5};
    for (int i = 0; i < nums.length; i++) {
      unit.insert(nums[i]);
    }

    for (int n : nums) {
      assertEquals(n, unit.worst().intValue());
      unit.removeWorst();
    }
  }

  @Test
  @DisplayName("[0.5] PQ removeBest() works: non-default Comparator")
  public void insertRemoveBestReverseOrder() {
    unit = createUnit(Collections.reverseOrder());

    int[] nums = {5, 4, 3, 2, 1, 0};
    for (int n : nums) {
      unit.insert(n);
    }

    for (int i = nums.length - 1; i >= 0; i--) {
      assertEquals(nums[i], unit.best().intValue());
      unit.removeBest();
    }
  }

  @Test
  @DisplayName("[0.5] PQ removeWorst() works: non-default Comparator")
  public void insertRemoveWorstReverseOrder() {
    unit = createUnit(Collections.reverseOrder());

    int[] nums = {5, 4, 3, 2, 1, 0};
    for (int n : nums) {
      unit.insert(n);
    }

    for (int i = 0; i < nums.length; i++) {
      assertEquals(nums[i], unit.worst().intValue());
      unit.removeWorst();
    }
  }

  @Test
  @DisplayName("[1] PQ operations work with duplicate elements")
  public void equalChildrenRemove() {
    int[] nums = {5, 3, 3, 2, 2, 0};
    for (int num : nums) {
      unit.insert(num);
    }

    unit.removeBest();
    assertEquals(3, unit.best().intValue());
    unit.removeWorst();
    assertEquals(2, unit.worst().intValue());
  }

  @Test
  @DisplayName("[1] PQ operations work (complex test)")
  public void complexTest() {
    int[] nums = {5, 5, 4, 3, 2, 2, 1, 0, 0};
    for (int num : nums) {
      unit.insert(num);
    }

    assertEquals(5, unit.best().intValue());
    assertEquals(0, unit.worst().intValue());
    unit.removeBest();
    assertEquals(5, unit.best().intValue());
    assertEquals(0, unit.worst().intValue());
    unit.removeWorst();
    assertEquals(5, unit.best().intValue());
    assertEquals(0, unit.worst().intValue());
    unit.removeBest();
    assertEquals(4, unit.best().intValue());
    assertEquals(0, unit.worst().intValue());
    unit.removeWorst();
    assertEquals(4, unit.best().intValue());
    assertEquals(1, unit.worst().intValue());
    unit.removeBest();
    assertEquals(3, unit.best().intValue());
    assertEquals(1, unit.worst().intValue());
    unit.removeWorst();
    assertEquals(3, unit.best().intValue());
    assertEquals(2, unit.worst().intValue());
    unit.removeBest();
    assertEquals(2, unit.best().intValue());
    assertEquals(2, unit.worst().intValue());
    unit.removeWorst();
    assertEquals(2, unit.best().intValue());
    assertEquals(2, unit.worst().intValue());

    if (Coin.flip().equals(Coin.HEAD)) {
      unit.removeBest();
    } else {
      unit.removeWorst();
    }

    assertThrows(EmptyException.class, () -> unit.best());
    assertThrows(EmptyException.class, () -> unit.worst());
    assertTrue(unit.empty());
  }
}
